from os import environ
from typing import List, Union

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport, BasicAuth

from saucelabs_visual.regions import Region
from saucelabs_visual.typing import IgnoreRegion, FullPageConfig


class SauceLabsVisual:
    client: Client = None
    build_id: Union[str, None] = None
    meta_cache: dict = {}

    def __init__(self):
        self._create_client()

    def _create_client(self):
        username = environ.get("SAUCE_USERNAME")
        access_key = environ.get("SAUCE_ACCESS_KEY")

        if username is None or access_key is None:
            raise Exception(
                'Sauce Labs credentials not set. Please check that you set correctly your '
                '`SAUCE_USERNAME` and `SAUCE_ACCESS_KEY` environment variables.'
            )

        region_url = Region.from_name(environ.get("SAUCE_REGION") or 'us-west-1').graphql_endpoint
        transport = AIOHTTPTransport(url=region_url, auth=BasicAuth(username, access_key))
        self.client = Client(transport=transport, execute_timeout=90)

    def get_client(self) -> Client:
        return self.client

    def create_build(
            self,
            name: str = environ.get('SAUCE_VISUAL_BUILD_NAME') or None,
            project: str = environ.get('SAUCE_VISUAL_PROJECT') or None,
            branch: str = environ.get('SAUCE_VISUAL_BRANCH') or None,
            default_branch: str = environ.get('SAUCE_VISUAL_DEFAULT_BRANCH') or None,
            custom_id: str = environ.get('SAUCE_VISUAL_CUSTOM_ID') or None,
            keep_alive_timeout: int = None
    ):
        query = gql(
            # language=GraphQL
            """
            mutation createBuild(
                $name: String!,
                $project: String,
                $branch: String,
                $customId: String,
                $defaultBranch: String,
                $keepAliveTimeout: Int,
            ) {
                createBuild(input: {
                    name: $name,
                    project: $project,
                    branch: $branch,
                    customId: $customId,
                    defaultBranch: $defaultBranch,
                    keepAliveTimeout: $keepAliveTimeout,
                }){
                    id
                    url
                }
            }
            """
        )
        values = {
            "name": name,
            "project": project,
            "branch": branch,
            "customId": custom_id,
            "defaultBranch": default_branch,
            "keepAliveTimeout": keep_alive_timeout,
        }
        build = self.client.execute(query, variable_values=values)
        self.build_id = build['createBuild']['id']
        return build

    def finish_build(self):
        query = gql(
            # language=GraphQL
            """
            mutation finishBuild($id: UUID!) {
                finishBuild(input: {uuid: $id}){
                    id
                    url
                    status
                }
            }
            """
        )
        values = {"id": self.build_id}
        self.meta_cache.clear()
        self.build_id = None
        return self.client.execute(query, variable_values=values)

    def get_selenium_metadata(self, session_id: str) -> str:
        query = gql(
            # language=GraphQL
            """
            query webdriverSessionInfo(
                $jobId: ID!,
                $sessionId: ID!
            ) {
                meta: webdriverSessionInfo(input: {
                    sessionId: $sessionId,
                    jobId: $jobId,
                }){
                    blob
                    operatingSystem
                    operatingSystemVersion
                    browser
                    browserVersion
                    deviceName
                }
            }
            """
        )
        values = {"sessionId": session_id, "jobId": session_id}
        meta = self.client.execute(query, variable_values=values)
        return meta['meta']['blob']

    def _get_meta(self, session_id: str) -> str:
        meta = self.meta_cache.get(session_id)

        if meta is None:
            meta = self.get_selenium_metadata(session_id)
            self.meta_cache[session_id] = meta

        return meta

    def create_snapshot_from_webdriver(
            self,
            name: str,
            session_id: str,
            test_name: str = None,
            suite_name: str = None,
            capture_dom: bool = False,
            clip_selector: str = None,
            ignore_regions: List[IgnoreRegion] = None,
            full_page_config: FullPageConfig = None,
    ):
        query = gql(
            # language=GraphQL
            """
            mutation createSnapshot(
                $name: String!,
                $sessionId: ID!,
                $meta: WebdriverSessionBlob!,
                $buildId: UUID!,
                $testName: String,
                $suiteName: String,
                $captureDom: Boolean,
                $clipSelector: String,
                $ignoreRegions: [RegionIn!],
                $fullPageConfig: FullPageConfigIn,
            ) {
                createSnapshotFromWebDriver(input: {
                    name: $name,
                    sessionId: $sessionId,
                    sessionMetadata: $meta,
                    buildUuid: $buildId,
                    testName: $testName,
                    suiteName: $suiteName,
                    captureDom: $captureDom,
                    clipSelector: $clipSelector,
                    ignoreRegions: $ignoreRegions,
                    fullPageConfig: $fullPageConfig,
                }){
                    id
                }
            }
            """
        )
        meta = self._get_meta(session_id)
        values = {
            "name": name,
            "sessionId": session_id,
            "meta": meta,
            "buildId": self.build_id,
            "testName": test_name,
            "suiteName": suite_name,
            "captureDom": capture_dom,
            "clipSelector": clip_selector,
            "ignoreRegions": ignore_regions,
            "fullPageConfig": {
                "delayAfterScrollMs": full_page_config.get('delay_after_scroll_ms'),
                "hideAfterFirstScroll": full_page_config.get('hide_after_first_scroll'),
            } if full_page_config is not None else None,
        }
        return self.client.execute(query, variable_values=values)



from typing import TypedDict, List, Optional


class IgnoreRegion(TypedDict):
    x: int
    y: int
    height: int
    width: int


class FullPageConfig(TypedDict):
    delay_after_scroll_ms: Optional[int]
    hide_after_first_scroll: Optional[List[str]]



from typing import Union, List


class Region:
    name: str
    aliases: List[str] = []
    graphql_endpoint: str
    job_url_template: str

    def __init__(
            self,
            name: str,
            aliases: List[str],
            graphql_endpoint: str,
            job_url_template: str
    ):
        self.name = name
        self.aliases = aliases
        self.graphql_endpoint = graphql_endpoint
        self.job_url_template = job_url_template

    def job_url(self, job_id: str) -> str:
        return self.job_url_template.format(JOB_ID=job_id)

    @staticmethod
    def from_name(region_name: str) -> 'Region':
        region: Union[Region, None] = None

        for i, v in enumerate(regions):
            if v.name == region_name or region_name in v.aliases:
                region = v
                break

        if region is None:
            raise Exception(f'"{region_name}" is not recognized as a Sauce Labs region.')

        return region


regions = [
    Region(
        name="staging",
        aliases=["us-west-4-jeh6"],
        graphql_endpoint='https://api.staging.saucelabs.net/v1/visual/graphql',
        job_url_template='https://app.staging.saucelabs.net/tests/{JOB_ID}',
    ),
    Region(
        name="us-east-4",
        aliases=["us-east-4-cm5i"],
        graphql_endpoint='https://api.us-east-4.saucelabs.com/v1/visual/graphql',
        job_url_template='https://app.us-east-4.saucelabs.com/tests/{JOB_ID}',
    ),
    Region(
        name="eu-central-1",
        aliases=["eu", "eu-west-3-lnbf"],
        graphql_endpoint='https://api.eu-central-1.saucelabs.com/v1/visual/graphql',
        job_url_template='https://app.eu-central-1.saucelabs.com/tests/{JOB_ID}',
    ),
    Region(
        name="us-west-1",
        aliases=['us', 'us-west-4-i3er'],
        graphql_endpoint='https://api.us-west-1.saucelabs.com/v1/visual/graphql',
        job_url_template='https://app.saucelabs.com/tests/{JOB_ID}',
    ),
]



from ast import literal_eval
from typing import List, Union

from SeleniumLibrary import SeleniumLibrary
from robot.api.deco import library, keyword
from robot.libraries.BuiltIn import BuiltIn

from saucelabs_visual.client import SauceLabsVisual as Client
from saucelabs_visual.typing import IgnoreRegion, FullPageConfig


@library(scope='GLOBAL')
class SauceLabsVisual:
    client: Client = None

    def __init__(self):
        self.client = Client()

    def _get_selenium_library(self) -> SeleniumLibrary:
        return BuiltIn().get_library_instance('SeleniumLibrary')

    def _get_selenium_id(self) -> str:
        return self._get_selenium_library().get_session_id()

    def _parse_full_page_config(self, value: str) -> Union[FullPageConfig, None]:
        """
        Parses the value from the 'Visual Snapshot' keyword to allow optional values and defaults
        that make sense. Otherwise, they would have to supply all properties to full_page_config
        which could get verbose.
        :param value:
        :return:
        """
        literal = literal_eval(value)
        literal_type = type(literal)
        parsed_value = None

        if literal_type is dict:
            parsed_value = literal
        elif literal_type is bool:
            parsed_value = {} if literal is True else None

        return FullPageConfig(
            delay_after_scroll_ms=parsed_value.get('delay_after_scroll_ms'),
            hide_after_first_scroll=parsed_value.get('delay_after_scroll_ms'),
        ) if parsed_value is not None else None

    @keyword(name="Create Visual Build")
    def create_visual_build(
            # Params 'duplicated' here, so we get type casting and named parameters provided by
            # Robot Framework for free.
            self,
            name: str = None,
            project: str = None,
            branch: str = None,
            default_branch: str = None,
            custom_id: str = None,
            keep_alive_timeout: int = None
    ):
        return self.client.create_build(
            name=name,
            project=project,
            branch=branch,
            default_branch=default_branch,
            custom_id=custom_id,
            keep_alive_timeout=keep_alive_timeout,
        )

    @keyword(name="Finish Visual Build")
    def finish_visual_build(self):
        return self.client.finish_build()

    @keyword(name="Visual Snapshot")
    def visual_snapshot(
            # Params 'duplicated' here, so we get type casting and named parameters provided by
            # Robot Framework for free.
            self,
            name: str,
            capture_dom: bool = False,
            clip_selector: str = None,
            ignore_regions: List[IgnoreRegion] = None,
            full_page_config: str = None,
    ):
        session_id = self._get_selenium_id()

        # Robot fails when attempting to parse a TypedDict out of a Union -- and converters are not
        # triggered. So, allow the default value as a string then parse it ourselves to allow us
        # to set proper default / optional values.
        parsed_fpc = self._parse_full_page_config(full_page_config) if (
                    type(full_page_config) is str) else None

        return self.client.create_snapshot_from_webdriver(
            name=name,
            session_id=session_id,
            test_name=BuiltIn().get_variable_value('\${TEST NAME}'),
            suite_name=BuiltIn().get_variable_value('\${SUITE NAME}'),
            capture_dom=capture_dom,
            clip_selector=clip_selector,
            ignore_regions=ignore_regions,
            full_page_config=parsed_fpc,
        )
